import torch
import torch.autograd
import torch.optim as optim
import torch.nn as nn
from torch.autograd import Variable
from newmodels import Critic, Actor
from newutils import Memory
import numpy as np







class DDPGagent:
    def __init__(self, env, droprob, l2penalty, hidden_size=256, actor_learning_rate=1e-4, critic_learning_rate=1e-3, gamma=0.99, tau=1e-2, max_memory_size=10000):
        # Seed
        seed = 0
        torch.manual_seed(seed)            # CPU seed
        torch.cuda.manual_seed(seed)       # current GPU seed
        torch.cuda.manual_seed_all(seed)   # all GPU seed
        
        # Params
        self.num_states = env.observation_space.shape[0]
        self.num_actions = env.action_space.shape[0]
        self.gamma = gamma
        self.tau = tau
        self.droprob = droprob
        self.l2penalty = l2penalty
        
        # Networks
        self.actor = Actor(self.num_states, hidden_size, self.num_actions, self.droprob, name='actor')
        self.actor_target = Actor(self.num_states, hidden_size, self.num_actions, self.droprob, name='actor_target')
        self.critic = Critic(self.num_states + self.num_actions, hidden_size, self.num_actions, self.droprob, name='critic')
        self.critic_target = Critic(self.num_states + self.num_actions, hidden_size, self.num_actions, self.droprob, name='critic_target')

        for target_param, param in zip(self.actor_target.parameters(), self.actor.parameters()):
            target_param.data.copy_(param.data)

        for target_param, param in zip(self.critic_target.parameters(), self.critic.parameters()):
            target_param.data.copy_(param.data)
        
        # Training
        self.memory = Memory(max_memory_size)        
        self.critic_criterion  = nn.MSELoss()
        self.actor_optimizer  = optim.Adam(self.actor.parameters(), lr=actor_learning_rate, weight_decay=self.l2penalty)
        self.critic_optimizer = optim.Adam(self.critic.parameters(), lr=critic_learning_rate, weight_decay=self.l2penalty)
    
    def get_action(self, state):
        state = Variable(torch.from_numpy(np.array(state)).float().unsqueeze(0)).to(self.actor.device)
        action = self.actor.forward(state).to(self.actor.device)
        action = action.detach().cpu().numpy()[0,0]
        return action
    
    def update(self, batch_size):
        states, actions, rewards, next_states, _ = self.memory.sample(batch_size)
        states = torch.FloatTensor(states).to(self.actor.device)
        actions = torch.FloatTensor(actions).to(self.actor.device)
        rewards = torch.FloatTensor(rewards).to(self.actor.device)
        next_states = torch.FloatTensor(next_states).to(self.actor.device)
    
        # Critic loss        
        Qvals = self.critic.forward(states, actions)
        next_actions = self.actor_target.forward(next_states)
        next_Q = self.critic_target.forward(next_states, next_actions.detach())
        Qprime = rewards + self.gamma * next_Q
        critic_loss = self.critic_criterion(Qvals, Qprime)
        
        # Actor loss
        policy_loss = -self.critic.forward(states, self.actor.forward(states)).mean()
        
        # update Actor networks
        self.actor_optimizer.zero_grad()
        policy_loss.backward()
        self.actor_optimizer.step()
        
        # update Critic networks
        self.critic_optimizer.zero_grad()
        critic_loss.backward() 
        self.critic_optimizer.step()

        # update target networks 
        for target_param, param in zip(self.actor_target.parameters(), self.actor.parameters()):
            target_param.data.copy_(param.data * self.tau + target_param.data * (1.0 - self.tau))
       
        for target_param, param in zip(self.critic_target.parameters(), self.critic.parameters()):
            target_param.data.copy_(param.data * self.tau + target_param.data * (1.0 - self.tau))
    
    def save_models(self):
        self.actor.save_checkpoint()
        self.actor_target.save_checkpoint()
        self.critic.save_checkpoint()
        self.critic_target.save_checkpoint()
        
    def load_models(self):
        self.actor.load_checkpoint()
        self.actor_target.load_checkpoint()
        self.critic.load_checkpoint()
        self.critic_target.load_checkpoint() 
    
    