# %% imports
import itertools
import os
import sys
import platform

if 'SUMO_HOME' in os.environ:
    tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
    sys.path.append(tools)
else:
    sys.exit("please declare environment variable 'SUMO_HOME'")
import traci
import traci.constants as tc
import xml.etree.ElementTree as ET
import pandas as pd
import copy
import numpy as np
import time
import math
from pylab import sin, cos, exp
from ast import literal_eval
import random
from shutil import copyfile
from gym import spaces
from skfuzzy import membership as mem


# %% global constants
class Constants():
    MAINDIREC = r"E:\CoMoVe\CoMoVe_RL20210308_TD3\CoMoVe_RL"
    VDMODEL = os.path.join(MAINDIREC, r"mathWorksFiles\simulinkFiles\VDModel")
    SUMOBINARY = r"E:\Program Files (x86)\Eclipse\Sumo\bin\sumo-gui" # sumo-gui for GUI
    FILENM = "simpleRoad"
    OSNAME = platform.system()
    RADARPENPCT = 100  # % of vehicles to have radar sensor
    INPUTPORT = ".InputPort(1).Data"
    INPUTPORT_SHORT = ".InputPort"  # for multiple calls
    OUTPUTPORT = ".OutputPort(1).Data"
    OUTPUTPORT_SHORT = ".OutputPort"  # for multiple calls
    SIMULINKMODELS = ["objectDetectionModel", "vdbasemodel"]
    SIMCMD = "SimulationCommand"
    MATLABCOMMANDS = ['start', 'pause', 'continue', 'step', 'update', 'stop']
    CMDSETUP = "set_param('{}','{}','{}')"
    VDSUBMODELS = ["Driveline", "PassVeh7DOF", "PassVeh14DOF"]
    BASESTRINGS = [['vdbasemodel/VDModel{}/PosX', "posX_{}"],
                   ['vdbasemodel/VDModel{}/PosY', "posY_{}"],
                   ['vdbasemodel/VDModel{}/Slip', "Slip_{}"],
                   ['vdbasemodel/VDModel{}/Environment/mu', "mu_{}"],
                   ['vdbasemodel/VDModel{}/Acceleration', "acc_{}"],
                   ['vdbasemodel/VDModel{}/Velocity', "vel_{}"],
                   ['objectDetectionModel/detectionModel_{}/out_radar', "radar_{}"],
                   ['objectDetectionModel/detectionModel_{}/out_camera', "camera_{}"],
                   ['vdbasemodel/VDModel{}/FullSlip', "FullSlip_{}"],
                   ['vdbasemodel/VDModel{}/AccelerationMS2', "accmss_{}"]
                   ]
    VUT = "veh0"  # vehicle under test
    VIR_VEH = "virtual"  # virtual Vehicle
    VEL_BOUND = [0, 130 / 3.6]  # velocity range 0-130kmph
    HW_BOUND = [0, 10]
    HWDELTA_BOUND = [-0.1, 0.1]
    STEPSPERACTION = 5
    WGHTSFORSTATEAGG = np.array([i + 1 for i in range(STEPSPERACTION)])


# %% SUMO Related Class
class SumoObject(Constants):
    def __init__(self):
        self.sumoCmd = [Constants.SUMOBINARY, "-c",
                        os.path.join(Constants.MAINDIREC, "sumoFiles/" + Constants.FILENM + ".sumocfg"),
                        "--start",
                        "--quit-on-end", "--no-step-log", "true", "-W", "true", "--duration-log.disable"]
        self.sumoVehicleDetails = {}  # like vehicle count, vehicle ID
        self.sumoVehicleData = {}  # current simulation vehicle data
        # offset for position conversion, node for road values, edge for edges, route for defined routes
        self.sumoExtensions = [".net", ".nod", ".edg", ".rou"]
        self.sumoNetworkValues = {"sumoOffset": [], "nodeValues": [], "edgeValues": [], "routeValues": []}

    def readXMLFiles(self, fileType):
        """
        :param fileType:sumo file extension
        To get the attributes from SUMO XML file to recreate the road network in MATLAB for RADAR sensor environment
        :return: Attributes
        """
        tree = ET.ElementTree(
            file=os.path.join(Constants.MAINDIREC, "sumoFiles/" + Constants.FILENM + fileType + ".xml"))
        return (list(tree.getroot()))

    def getNetworkValues(self, extensions):
        """
        :param extensions:types of sumo extensions. check the init function
        :return: no return. Save the values in class variables to access them in another class
        """
        # sumoOffset
        attributes = self.readXMLFiles(extensions[0])
        idx = [idx for idx, i in enumerate(attributes) if 'location' in i.tag][0]
        self.sumoNetworkValues["sumoOffset"] = [float(i) for i in attributes[idx].attrib["netOffset"].split(",")]
        # node values
        self.sumoNetworkValues["nodeValues"] = \
            [i.attrib for i in self.readXMLFiles(extensions[1]) if 'node' in i.tag]
        # edges between nodes
        self.sumoNetworkValues["edgeValues"] = \
            [i.attrib for i in self.readXMLFiles(extensions[2]) if 'edge' in i.tag]
        # route generated between edges
        routesFile = self.readXMLFiles(extensions[3])
        self.sumoNetworkValues["routeValues"] = \
            [i.attrib for i in routesFile if 'route' in i.tag]

        # vehicle details
        self.sumoVehicleDetails["vehiclesType"] = \
            [i.attrib for i in routesFile if 'vType' in i.tag]
        self.sumoVehicleDetails["vehicles"] = \
            [i.attrib for i in routesFile if 'vehicle' in i.tag]
        self.sumoVehicleDetails["vehiclesID"] = {i["id"]: idx for idx, i in
                                                 enumerate(self.sumoVehicleDetails["vehicles"], 1)}
        self.sumoVehicleDetails["numOfVehicles"] = len(self.sumoVehicleDetails["vehiclesID"].keys())

    def getInitialValues(self, sumoNetvalues, sumoVehicles):
        """
        :param sumoNetvalues: SUMO road network nodes, edges, routes.
        :param vehicles: vehicles ID from SUMO
        For creating the Actors in MATLAB driving scenario designer. Mandatory for RADAR and CAMERA sensor in simulink
        models
        :return:no Return. save them in the class variables
        """
        vehicles = sumoVehicles["vehicles"]
        vehiclesType = sumoVehicles["vehiclesType"]
        nodeValues, edgeValues = sumoNetvalues["nodeValues"], sumoNetvalues["edgeValues"]
        routeValues = sumoNetvalues["routeValues"]
        for route in routeValues:
            routeVal = [[i['id'], i["departSpeed"], i.get("departPos")] for i in vehicles if i["route"] == route["id"]]
            for i in routeVal:
                vehID, speed, departPos = i
                departPos = [0 if departPos is None else float(departPos)][0]
                self.sumoVehicleData[vehID] = {}
                # vehicle type max speed is considered as its desired Velocity
                self.sumoVehicleData[vehID]["desiredVel"] = float([i["maxSpeed"] for i in vehiclesType
                                                                   for j in vehicles
                                                                   if i["id"] == j["type"] and j["id"] == vehID][0])
                [frm, to] = [[i["from"], i["to"]] for i in edgeValues if i["id"] == route["edges"].split(" ")[0]][0]
                [orgx, orgy] = [[float(i["x"]), float(i["y"])] for i in nodeValues if i["id"] == frm][0]
                [destx, desty] = [[float(i["x"]), float(i["y"])] for i in nodeValues if i["id"] == to][0]
                if orgy == desty:
                    if destx - orgx > 0:
                        latPos = -1.65
                    else:
                        latPos = 1.65
                else:
                    if desty - orgy > 0:
                        latPos = 1.65
                    else:
                        latPos = -1.65
                self.sumoVehicleData[vehID]["speed"] = float(speed)
                self.sumoVehicleData[vehID]["speedVector"] = [float(speed), 0, 0]
                self.sumoVehicleData[vehID]["pos"] = [float(orgx) + departPos, float(latPos), 0]
                self.sumoVehicleData[vehID]["yaw"] = 0
                self.sumoVehicleData[vehID]["speedChange"] = True
                self.sumoVehicleData[vehID]["laneChange"] = True
                self.sumoVehicleData[vehID]["numID"] = self.sumoVehicleDetails["vehiclesID"][vehID]
                # useful for reinforcement learning reset method
                self.sumoVehicleData[vehID]["initialPosition"] = [float(orgx) + departPos, float(latPos), 0]
                self.sumoVehicleData[vehID]["initialSpeed"] = float(speed)


# %% MATLAB and Simulink related class
class VehicleModel(Constants):
    def __init__(self, initial):
        if initial:
            import matlab.engine as engine
            if engine.find_matlab():
                self.eng = engine.connect_matlab(engine.find_matlab()[0])  # engine.find_matlab()[0]
            else:
                sys.exit("no Matlab engine shared")
            self.eng.eval("cd {}\mathWorksFiles;"
                          "addpath('{}\mathWorksFiles\matlabFiles');"
                          "addpath('{}\mathWorksFiles\simulinkFiles')".
                          format(Constants.MAINDIREC, Constants.MAINDIREC, Constants.MAINDIREC), nargout=0)
            self.eng.eval("open_system('objectDetectionModel');open_system('baseDetectionModel')", nargout=0)
            self.radarMaxDet = 0
            self.cameraMaxDet = 0

    def initDetectionModels(self, radarVehID, vehDetectionInfo, vehiclesID):
        """
        :param radarVehID: vehicles that equipped with RADAR
        :param vehDetectionInfo: initial vehicle position, speed values from SUMO
        :param vehiclesID: vehicles IDs from SUMO
        :return: no Return.
        """
        for curveh in radarVehID:
            relDis = []
            relVel = []
            yaw = []
            actorIDs = []
            curPos = vehDetectionInfo[curveh]["pos"]
            curVel = vehDetectionInfo[curveh]["speedVector"]
            for leadveh in vehiclesID:
                if curveh != leadveh:
                    vehyaw = 90
                    actorIDs.append(vehiclesID[leadveh])
                    relDis.append(list(np.array(vehDetectionInfo[leadveh]["pos"]) - np.array(curPos)))
                    relVel.append(list(np.array(vehDetectionInfo[leadveh]["speedVector"]) - np.array(curVel)))
                    yaw.append(vehyaw)
            vehDetectionInfo[curveh]["actorOrder"] = actorIDs
            self.eng.eval(
                '[curcar, car_bus, x_] = envCreateActors('
                '{actorIDs},{curVehID},{relDist},{relVel},{yaw},"create");'
                'vehData.("{carID}") = curcar;carBuses.("{buscarID}") = '
                'car_bus;car_{savecarID} = curcar;'.format(
                    actorIDs=actorIDs, curVehID=radarVehID[curveh], relDist=relDis, relVel=relVel, yaw=yaw,
                    carID="car_" + str(radarVehID[curveh]), buscarID="car_" + str(radarVehID[curveh]),
                    savecarID=str(radarVehID[curveh])),
                nargout=0)

        self.radarMaxDet = int(self.eng.eval('get_param("baseDetectionModel/Radar_Camera_Detections/'
                                             'Radar Detection Generator", "MaxNumDetections")', nargout=1))
        self.cameraMaxDet = int(self.eng.eval('get_param("baseDetectionModel/Radar_Camera_Detections/'
                                              'Vision Detection Generator", "MaxNumDetections")', nargout=1))
        self.eng.eval("updateDetectionModel({},carBuses,'add')".format(list(radarVehID.values())), nargout=0)

    def createVDModels(self, vehCnt, vehDetectionInfo):
        """
        :param vehCnt: number of vehicles involved in the simulation
        :param vehDetectionInfo: initial vehicle values from SUMO
        For creating the vehicle dynamics model for each vehicle in Simulink
        :return: no return
        """
        '''
        for model in Constants.VDSUBMODELS:
            for i in range(1, vehCnt + 1):
                copyfile(Constants.VDMODEL + r"\refs\{}\Plant\{}.slx".format(model, model),
                         Constants.VDMODEL + r"\refs\{}\Plant\{}.slx".format(model, model + "_" + str(i)))
        self.eng.eval("updateVDModel({},'add')".format(vehCnt), nargout=0)
        '''
        self.eng.eval("Init;updateVDModel_new({},'add')".format(vehCnt), nargout=0)
        vehData = [[vehDetectionInfo[i]["pos"][:2], vehDetectionInfo[i]["initialSpeed"], vehDetectionInfo[i]["numID"]]
                   for i in vehDetectionInfo]
        # self.eng.eval("vehInitialData({},{})".format(vehCnt, vehData), nargout=0)
        self.eng.eval("VEHICLES = vehInitialData_NEW({},{})".format(vehCnt, vehData), nargout=0)

    def deleteVDModels(self, vehCnt):
        """
        :param vehCnt: number of vehicles involved in the simulation
        delete the vehicle dynamics blocks in the Simulink mode
        :return: no return
        """
        checkCmd = "get_param('" + Constants.SIMULINKMODELS[1] + "','SimulationStatus')"
        exec_status = self.eng.eval(checkCmd, nargout=1)
        # check for simulink simulation status
        if exec_status != "stopped":
            cmd = Constants.CMDSETUP.format(
                Constants.SIMULINKMODELS[1], Constants.SIMCMD, Constants.MATLABCOMMANDS[5])
            self.eng.eval(cmd, nargout=0)
        '''
        for model in Constants.VDSUBMODELS:
            for i in range(1, vehCnt + 1):
                os.remove(Constants.VDMODEL + r"\refs\{}\Plant\{}.slx".format(model, model + "_" + str(i)))

        self.eng.eval("updateVDModel({},'delete')".format(vehCnt), nargout=0)
        '''
        self.eng.eval("updateVDModel_new({},'delete')".format(vehCnt), nargout=0)

    def deleteDetectionsModels(self, radarVehID):
        """
        :param radarVehID: vehicles equipped with RADAR
        delete the vehicles from Object detection SIMULINK model
        :return: no return
        """
        checkCmd = "get_param('" + Constants.SIMULINKMODELS[0] + "','SimulationStatus')"
        exec_status = self.eng.eval(checkCmd, nargout=1)
        # check for simulink simulation status
        if exec_status != "stopped":
            cmd = Constants.CMDSETUP.format(
                Constants.SIMULINKMODELS[0], Constants.SIMCMD, Constants.MATLABCOMMANDS[5])
            self.eng.eval(cmd, nargout=0)
        self.eng.eval("updateDetectionModel({},'','delete')".format(radarVehID), nargout=0)


# %% simulation
class SimulationEnvironment(Constants):
    def __init__(self):
        self.sumoInfo = None
        self.vehCnt = None
        self.vehiclesID = None
        self.vehDetectionInfo = None
        self.vdObject = None
        self.initializeSimulation()
        self.simuGetData = {}
        self.matlabEng = self.vdObject.eng
        self.arrivedVehicles = []
        self.sensorValues = []
        self.simulationVehIDs = []
        self.accNewVel = {}
        self.VUT = Constants.VUT
        self.done = False  # A flag to check if goal is reached.
        self.zeroVelCounter = 0
        self.timeStep = 0.1
        self.traciInitial = True
        self.initialStep = True
        self.blkCmds = ["vdbasemodel/VDModel{}/LngVelRef", "vdbasemodel/VDModel{}/LtrlPosRef",
                        'vdbasemodel/VDModel{}/Environment/FrictionFL',
                        'vdbasemodel/VDModel{}/Environment/FrictionRL', "vdbasemodel/VDModel{}/desiredAcc",
                        'vdbasemodel/VDModel{}/Passenger Vehicle/Driveline/brkSwitch',
                        'vdbasemodel/VDModel{}/Controllers/switchTrq',
                        'vdbasemodel/VDModel{}/Controllers/switchAccelCmd']
        self.sensorVehDistances = []
        self.splitUp = {}
        self.splitUpNoJerk = {}
        self.simulationTime = 0.0
        self.fricChange = ["ones(1, 1).*0.35", "ones(1, 1).*1.0"]
        self.fricChangePos = [105, 330]
        self.fricChangeInd = {i: [[0, 0], [0, 0]] for i in list(self.vehiclesID.keys())}

        tableNoJerk = list(itertools.product([False, True], repeat=2))
        weightageNoJerk = [[1 / 2, 1 / 2], [2 / 3, 1 / 3], [1 / 3, 2 / 3], [1 / 2, 1 / 2]]
        for i, j in zip(tableNoJerk, weightageNoJerk):
            self.splitUpNoJerk[i] = j

        table = list(itertools.product([False, True], repeat=3))
        weightage = [[1 / 3, 1 / 3, 1 / 3], [2.5 / 6, 2.5 / 6, 1 / 6], [2.5 / 6, 1 / 6, 2.5 / 6], [2 / 3, 1 / 6, 1 / 6],
                     [1 / 6, 2.5 / 6, 2.5 / 6], [1 / 6, 2 / 3, 1 / 6], [1 / 6, 1 / 6, 2 / 3], [1 / 3, 1 / 3, 1 / 3]]
        for i, j in zip(table, weightage):
            self.splitUp[i] = j

        self.actions = pd.DataFrame([], columns=["acceleration"])

        self.leadVehSpeeds = pd.read_csv(r"E:\CoMoVe\CoMoVe_RL20210308_TD3\CoMoVe_RL\pythonFiles\leadVehSpeeds_new.csv")
        # vehicle values
        self.egoVehX, self.egoVehY = None, None
        self.leadVehX, self.leadVehY = None, None
        self.egoVelocity = None
        self.leadVelocity = None
        self.desiredHW = 2.0
        self.egoDesiredSpeed = None
        self.virtVehVel = self.egoDesiredSpeed
        self.egoPrevAcc = 0
        self.egoVehAcc = 0
        self.bias = 0
        self.delta_distance = None
        self.delta_velocity = None

        # state & reward values
        self.hw, self.hwdelta = None, None
        self.preHW = 0
        self.prevVehAcc = 0
        self.jerk = 0
        self.ttc = None
        self.longSlip = None
        self.frictionMu = None
        self.fullSlip = None
        self.fullFriction = None
        # long slip n friction
        self.longSlip_FL = None
        self.longSlip_FR = None
        self.longSlip_RL = None
        self.longSlip_RR = None
        self.frictionMu_FL = None
        self.frictionMu_FR = None
        self.frictionMu_RL = None
        self.frictionMu_RR = None
        
        '''self.observation_space = np.array([self.prevVehAcc, self.hw, self.hwdelta,
                                           self.longSlip_FL, self.longSlip_FR, self.longSlip_RL, self.longSlip_RR,
                                           self.frictionMu_FL, self.frictionMu_FR, self.frictionMu_RL,
                                           self.frictionMu_RR,
                                           ])'''
    
        self.observation_space = np.array([self.egoVelocity, self.delta_distance, 
                                           self.delta_velocity, 
                                           self.longSlip_FL, self.longSlip_FR, 
                                           self.longSlip_RL, self.longSlip_RR,
                                           self.frictionMu_FL, self.frictionMu_FR, 
                                           self.frictionMu_RL, self.frictionMu_RR])
        
        self.action_space = spaces.Box(np.array([-2.0]), np.array([+1.47]), dtype=np.float32)

    def initializeSimulation(self):
        # Get road network, vehicle details from SUMO
        self.sumoInfo = SumoObject()
        self.sumoInfo.getNetworkValues(self.sumoInfo.sumoExtensions)
        self.sumoInfo.getInitialValues(self.sumoInfo.sumoNetworkValues, self.sumoInfo.sumoVehicleDetails)
        self.vehCnt = self.sumoInfo.sumoVehicleDetails["numOfVehicles"]
        self.vehiclesID = self.sumoInfo.sumoVehicleDetails["vehiclesID"]
        self.vehDetectionInfo = self.sumoInfo.sumoVehicleData

        # Choose the vehicles to have RADAR sensor
        radarlessveh = []
        noradarvehcnt = int(self.vehCnt - math.floor(self.vehCnt * (Constants.RADARPENPCT / 100.0)))
        while len(radarlessveh) < noradarvehcnt:
            x = random.choice(range(0, self.vehCnt))
            if "veh" + str(x) not in radarlessveh:
                radarlessveh.append("veh" + str(x))
        self.radarVehicles = {i: self.vehiclesID[i] for i in set(self.vehiclesID) - set(radarlessveh)}

        # initialise MATLAB and SIMULINK ObjectDetection and Vehicle Dynamics Models
        self.vdObject = VehicleModel(True)
        self.vdObject.initDetectionModels(self.radarVehicles, self.vehDetectionInfo, self.vehiclesID)
        self.vdObject.createVDModels(self.vehCnt, self.vehDetectionInfo)

    def findNearbyVehicle(self):
        dist = {}
        for i in self.vehDetectionInfo:
            if i != self.VUT and i != Constants.VIR_VEH:
                dist[i] = np.linalg.norm(np.array(self.vehDetectionInfo[self.VUT]["initialPosition"]) -
                                         np.array(self.vehDetectionInfo[i]["initialPosition"]))
        return (min(dist, key=dist.get))

    def reset(self):
        self.stopSimulation("reset")
        time.sleep(0.1)             
        
        self.actions = pd.DataFrame([], columns=["acceleration"])
        self.fricChangeInd = {i: [[0, 0], [0, 0]] for i in list(self.vehiclesID.keys())}
        self.frictionChange("", "", "", initial=1)
        self.simulationTime = 0.0
        self.traciInitial = True
        self.initialStep = True
        # update initial data (velocity, lateralpos )into models
        cmdExec = ""
        for val in self.vehiclesID:
            cmdExec = cmdExec + \
                      "set_param('{}', 'Value', '{}');".format(self.blkCmds[0].format(self.vehiclesID[val]),
                                                               str(self.vehDetectionInfo[val]["initialSpeed"] + 0.0))
            cmdExec = cmdExec + \
                      "set_param('{}', 'Value', '{}');".format(self.blkCmds[4].format(self.vehiclesID[val]),
                                                               str(0))
            cmdExec = cmdExec + \
                      "set_param('{}', 'Value', '{}');".format(self.blkCmds[1].format(self.vehiclesID[val]),
                                                               str(self.vehDetectionInfo[val]["initialPosition"][1]))
            cmdExec = cmdExec + \
                      "set_param('{}', 'Value', '{}');".format(self.blkCmds[5].format(self.vehiclesID[val]),
                                                               str(0))
            cmdExec = cmdExec + \
                      "set_param('{}', 'Value', '{}');".format(self.blkCmds[6].format(self.vehiclesID[val]),
                                                               str(0))
            cmdExec = cmdExec + \
                      "set_param('{}', 'Value', '{}');".format(self.blkCmds[7].format(self.vehiclesID[val]),
                                                               str(0))
        self.matlabEng.eval(cmdExec, nargout=0)
        self.egoVehX, self.egoVehY = self.vehDetectionInfo[self.VUT]["initialPosition"][:2]
        leadVehName = self.findNearbyVehicle()
        egoVehID, leadVehID = self.vehDetectionInfo[self.VUT]["numID"], self.vehDetectionInfo[leadVehName]["numID"]
        self.leadVehX, self.leadVehY = self.vehDetectionInfo[leadVehName]["initialPosition"][:2]
        self.egoVelocity = self.vehDetectionInfo[self.VUT]["initialSpeed"]
        self.leadVelocity = self.vehDetectionInfo[leadVehName]["initialSpeed"]
        self.desiredHW = 2.0
        # self.vehDetectionInfo[self.VUT]["desiredVel"] = 22.25
        self.egoDesiredSpeed = round(self.vehDetectionInfo[self.VUT]["desiredVel"], 2)
        self.done = False

        # creating runtime objects command
        exec_cmd = ""
        # start detection model
        exec_cmd = exec_cmd + Constants.CMDSETUP.format(
            Constants.SIMULINKMODELS[0], Constants.SIMCMD, Constants.MATLABCOMMANDS[0]) + ";"
        # VD Model
        exec_cmd = exec_cmd + Constants.CMDSETUP.format(
            Constants.SIMULINKMODELS[1], Constants.SIMCMD, Constants.MATLABCOMMANDS[0]) + ";"
        for vehID in self.vehiclesID:
            self.simuGetData[vehID] = {"blk": [], "rto": []}
            for j in Constants.BASESTRINGS:
                self.simuGetData[vehID]["blk"].append(j[0].format(str(self.vehiclesID[vehID])))
                self.simuGetData[vehID]["rto"].append(j[1].format(str(self.vehiclesID[vehID])))
                exec_cmd = exec_cmd + \
                           (self.simuGetData[vehID]["rto"][-1] +
                            "= get_param('" + self.simuGetData[vehID]["blk"][-1] + "','RuntimeObject');")

        print("starting the simulation")
        self.matlabEng.eval("open_system('objectDetectionModel');open_system('baseDetectionModel');clear;clc",
                            nargout=0)
        self.vdObject.initDetectionModels(self.radarVehicles, self.vehDetectionInfo, self.vehiclesID)
        vehData = [[self.vehDetectionInfo[i]["initialPosition"][:2], self.vehDetectionInfo[i]["initialSpeed"],
                    self.vehDetectionInfo[i]["numID"]]
                   for i in self.vehDetectionInfo]
        # self.eng.eval("vehInitialData({},{})".format(vehCnt, vehData), nargout=0)
        self.matlabEng.eval("Init;VEHICLES = vehInitialData_NEW({},{})".format(self.vehCnt, vehData), nargout=0)
        traci.start(self.sumoInfo.sumoCmd)
        self.traciObject = traci
        # self.matlabEng.eval(
        #     Constants.CMDSETUP.format(Constants.SIMULINKMODELS[0], Constants.SIMCMD, Constants.MATLABCOMMANDS[0]),
        #     nargout=0)
        # self.matlabEng.eval(
        #     Constants.CMDSETUP.format(Constants.SIMULINKMODELS[1], Constants.SIMCMD, Constants.MATLABCOMMANDS[0]),
        #     nargout=0)
        self.matlabEng.eval(exec_cmd, nargout=0)

        # waiting for VD models to get stabilized (5 seconds)
        vdModelStatus = self.matlabEng.eval("get_param('{}','SimulationStatus')".format(Constants.SIMULINKMODELS[1]),
                                            nargout=1)
        while vdModelStatus == "running":
            vdModelStatus = self.matlabEng.eval(
                "get_param('{}','SimulationStatus')".format(Constants.SIMULINKMODELS[1]),
                nargout=1)
            time.sleep(0.5)
        cmdExec = ""
        for val in self.vehiclesID:
            cmdExec = cmdExec + \
                      "set_param('{}', 'Value', '{}');".format(self.blkCmds[5].format(self.vehiclesID[val]),
                                                               str(1))
            cmdExec = cmdExec + \
                      "set_param('{}', 'Value', '{}');".format(self.blkCmds[6].format(self.vehiclesID[val]),
                                                               str(1))
            cmdExec = cmdExec + \
                      "set_param('{}', 'Value', '{}');".format(self.blkCmds[7].format(self.vehiclesID[val]),
                                                               str(1))
        self.matlabEng.eval(cmdExec, nargout=0)
        # assigning stable Longitudinal Vehicle position as intial pos in VD model
        for i in self.simuGetData:
            cmdToRun = self.simuGetData[i]["rto"][0] + Constants.INPUTPORT
            self.vehDetectionInfo[i]["vdInitialPosition"] = self.matlabEng.eval(cmdToRun, nargout=1) - \
                                                            self.vehDetectionInfo[i]["initialPosition"][0]

        self.updateVDData(leadVehName)
        
        self.vehDetectionInfo[self.VUT]["speed"] = self.egoVelocity
        self.vehDetectionInfo[leadVehName]["speed"] = self.leadVelocity
        # retrieve initial vehicle values
        self.hw, self.hwdelta = (self.leadVehX - self.egoVehX) / max(self.egoVelocity, 1), 0
        self.preHW = self.hw
        self.jerk = 0
        self.ttc = [(self.leadVehX - self.egoVehX) / (self.egoVelocity - self.leadVelocity)
                    if self.egoVelocity < self.leadVelocity else 10000][0]
        
        self.delta_distance = self.leadVehX - self.egoVehX
        self.delta_velocity = self.leadVelocity - self.egoVelocity

        self.updateVDData(leadVehName)
        # self.boundingStates()
        '''states = np.array([round(self.prevVehAcc, 4), round(self.hw, 4), round(self.hwdelta, 4),
                           round(self.longSlip_FL, 4), round(self.longSlip_FR, 4), round(self.longSlip_RL, 4),
                           round(self.longSlip_RR, 4),
                           round(self.frictionMu_FL, 4), round(self.frictionMu_FR, 4), round(self.frictionMu_RL, 4),
                           round(self.frictionMu_RR, 4)
                           ])'''
        
        states = np.array([round(self.egoVelocity, 4), round(self.delta_distance, 4), 
                           round(self.delta_velocity, 4), 
                           round(self.longSlip_FL, 4), round(self.longSlip_FR, 4), 
                           round(self.longSlip_RL, 4), round(self.longSlip_RR, 4),
                           round(self.frictionMu_FL, 4), round(self.frictionMu_FR, 4), 
                           round(self.frictionMu_RL, 4), round(self.frictionMu_RR, 4)])
        
        
        
        self.egoPrevAcc = round(self.egoVehAcc, 4)
        return (states)

    def nextStep(self):
        """
        Next step of the Simulation. Get the current vehicle information from SUMO and update it in the Simulink
        object detection model for RADAR and Camera sensor
        :return: simulation status
        """
        try:
            # check for SUMO simulation status
            if traci.simulation.getMinExpectedNumber() > 0:
                simStatus = 1
                if traci.simulation.getLoadedNumber() > 0:
                    self.simulationVehIDs.extend(traci.simulation.getLoadedIDList())
                    for j, i in enumerate(self.simulationVehIDs):
                        traci.vehicle.subscribe(i, (
                            tc.VAR_SPEED, tc.VAR_ANGLE, tc.VAR_ROAD_ID, tc.VAR_LANEPOSITION, tc.VAR_LANE_INDEX))
                traci.simulationStep()
                listOfVeh = list(traci.vehicle.getIDList())
                self.arrivedVehicles.extend([self.vehiclesID[i] for i in traci.simulation.getArrivedIDList()])
                if self.simulationVehIDs != listOfVeh:
                    del self.simulationVehIDs[:]
                    self.simulationVehIDs.extend(listOfVeh)
                    # setting speed & speed mode for SUMO vehicles to avoid following the safety rules imposed by SUMO
                    for i in listOfVeh:
                        traci.vehicle.setSpeed(i, self.vehDetectionInfo[i]["initialSpeed"])
                        traci.vehicle.setSpeedMode(i, 0)

                # all the simulated vehicles arrived
                if len(self.simulationVehIDs) == 0:
                    return ("All vehicles reached the destination", 0)

                # get the new vehicle information from SUMO
                for curVehID in self.simulationVehIDs:
                    vehStatus = traci.vehicle.getSubscriptionResults(curVehID)
                    [posx, posy, posz] = traci.simulation.convert3D \
                        (vehStatus[tc.VAR_ROAD_ID],
                         vehStatus[tc.VAR_LANEPOSITION],
                         vehStatus[tc.VAR_LANE_INDEX], 'False')
                    [posy, posx, posz] = [round(posy, 4), round(posx, 4), round(posz, 4)]
                    curVehAngle = vehStatus[tc.VAR_ANGLE]
                    directionVector = np.array(
                        [round(sin(math.radians(curVehAngle)), 2), round(cos(math.radians(curVehAngle)), 2), 0]) * \
                                      vehStatus[tc.VAR_SPEED]

                    self.vehDetectionInfo[curVehID]["speedVector"] = directionVector

                    if posy != self.vehDetectionInfo[curVehID]["pos"][1]:
                        self.vehDetectionInfo[curVehID]["laneChange"] = True
                    else:
                        self.vehDetectionInfo[curVehID]["laneChange"] = False

                    if vehStatus[tc.VAR_SPEED] != self.vehDetectionInfo[curVehID]["speed"]:
                        self.vehDetectionInfo[curVehID]["speedChange"] = True
                    else:
                        self.vehDetectionInfo[curVehID]["speedChange"] = False

                    self.vehDetectionInfo[curVehID]["speed"] = vehStatus[tc.VAR_SPEED]
                    self.vehDetectionInfo[curVehID]["pos"] = [posx, posy, posz]
                    self.vehDetectionInfo[curVehID]["yaw"] = curVehAngle

                # setting speed and speed mode for SUMO vehicles to avoid following the safety rules imposed by SUMO
                if self.traciInitial:
                    self.traciInitial = False
                    for i in self.simulationVehIDs:
                        traci.vehicle.setSpeed(i, self.vehDetectionInfo[i]["initialSpeed"])
                        traci.vehicle.setSpeedMode(i, 0)
                    # virutal vehicle if no other vehicles in Ego Vehicle RADAR range
                    self.vehDetectionInfo[Constants.VIR_VEH] = self.vehDetectionInfo[self.VUT].copy()
                    self.vehDetectionInfo[Constants.VIR_VEH]["active"] = False

                # updating current vehicle position in Simulink model
                for veh in self.radarVehicles:
                    if veh not in self.arrivedVehicles:
                        relDis = []
                        relVel = []
                        yaw = []
                        actorIDs = []
                        curPos = self.vehDetectionInfo[veh]["pos"]
                        curVel = self.vehDetectionInfo[veh]["speedVector"]
                        curYaw = self.vehDetectionInfo[veh]["yaw"]
                        for m in self.vehDetectionInfo[veh]["actorOrder"]:
                            leadVehId = [i[0] for i in list(self.vehiclesID.items()) if
                                         i[1] == m][0]
                            if leadVehId not in self.arrivedVehicles:
                                leadVehYaw = self.vehDetectionInfo[leadVehId]["yaw"]
                                actorIDs.append(m)
                                if curYaw == leadVehYaw:
                                    direcVector = np.array([math.cos(math.radians(0)), math.sin(math.radians(0)), 0])
                                    reldist = np.array(self.vehDetectionInfo[leadVehId]["pos"]) * direcVector - \
                                              np.array(curPos) * direcVector
                                    velDiff = np.array(self.vehDetectionInfo[leadVehId]["speedVector"]) * direcVector - \
                                              np.array(curVel) * direcVector
                                else:
                                    direcVector = np.array([math.cos(math.radians(leadVehYaw - curYaw)),
                                                            math.sin(math.radians(leadVehYaw - curYaw)), 0])
                                    reldist = np.array(self.vehDetectionInfo[leadVehId]["pos"]) * direcVector - \
                                              np.array(curPos)
                                    velDiff = np.array(self.vehDetectionInfo[leadVehId]["speedVector"]) * direcVector - \
                                              np.array(curVel)
                                relDis.append(list(reldist))
                                relVel.append(list(velDiff))
                                yaw.append(leadVehYaw)
                        curVehNum = str(self.radarVehicles[veh])
                        self.matlabEng.eval(
                            '[x_, y_, vehDataUpdate] = envCreateActors('
                            '{actorIDs},{curVehID},{relDist},{relVel},{yaw},"update",vehData.("{vehID}"));'
                            'vehData.car_{savestruct}=vehDataUpdate;car_{savevar} = vehData.car_{loadstruct};'
                            'set_param("objectDetectionModel/detectionModel_{mdlnum}/Radar_Camera_Detections/scenario"'
                            ',"Value","car_{varname}");'.
                                format(actorIDs=actorIDs, curVehID=curVehNum, relDist=relDis, relVel=relVel,
                                       yaw=yaw, vehID="car_" + curVehNum, savestruct=curVehNum,
                                       savevar=curVehNum, loadstruct=curVehNum,
                                       mdlnum=curVehNum, varname=curVehNum),
                            nargout=0)
                return ("_", simStatus)
            else:
                simStatus = 0
                # traci.close()
                return ("No Vehicle Info. Simulation Ended", simStatus)
        except Exception as e:
            # traci.close()
            print("Something went wrong: {}".format(e))
            return (e, 0)

    
    # def leadVehicleAcceleration(self, simTime): #Test scenario: also need to change ego and lead velocity = 17.5 in simpleRoad.rou.xml
    #     newVel = list(self.leadVehSpeeds[self.leadVehSpeeds["Time"] == round(simTime, 1)]["Speed"])[0]
    #     cmdExec = "set_param('{}', 'Value', '{}');". \
    #             format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
    #     self.matlabEng.eval(cmdExec, nargout=0)        
    

    def frictionChange(self, idx, vehicle, frictionValue, initial=0):
        # friction change
        if initial:
            cmdExec = ""
            for i in list(self.vehiclesID.keys()):
                cmdExec = cmdExec + "set_param('{}', 'Value', 'ones(1, 1).*1.0');". \
                    format(self.blkCmds[2].format(self.vehiclesID[i]))
                cmdExec = cmdExec + "set_param('{}', 'Value', 'ones(1, 1).*1.0');". \
                    format(self.blkCmds[3].format(self.vehiclesID[i]))
            if cmdExec:
                self.matlabEng.eval(cmdExec, nargout=0)
        else:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[idx].format(self.vehiclesID[vehicle]), frictionValue)
            self.matlabEng.eval(cmdExec, nargout=0)

    def leadVehicleAcceleration(self, simTime):
        if 0.0 <= simTime <= 5.0:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (0))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 5.0 <= simTime <= 5.5:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.25))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 5.5 <= simTime <= 6.0:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.25))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 6 <= simTime <= 6.5:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.2))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 6.5 <= simTime <= 7:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.2))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 6.6 <= simTime <= 7:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.15))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 7 <= simTime <= 7.4:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.15))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 7.4 <= simTime <= 7.8:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.1))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 7.8 <= simTime <= 8.2:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.1))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 8.2 <= simTime <= 8.6:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.05))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 8.6 <= simTime <= 9.0:
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (-0.05))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 8.2 <= simTime <= 13:
            newVel = 18 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 13 <= simTime <= 13.2:
            newVel = 18.2 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 13.2 <= simTime <= 13.4:
            newVel = 18.4 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 13.4 <= simTime <= 13.6:
            newVel = 18.6 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 13.6 <= simTime <= 13.8:
            newVel = 18.8 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 13.8 <= simTime <= 14:
            newVel = 19 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 14 <= simTime <= 14.2:
            newVel = 19.2 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 14.2 <= simTime <= 14.4:
            newVel = 19.4 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 14.4 <= simTime <= 14.6:
            newVel = 19.6 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        elif 14.6 <= simTime <= 14.8:
            newVel = 19.8 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)
        else:
            newVel = 20 + self.bias
            cmdExec = "set_param('{}', 'Value', '{}');". \
                format(self.blkCmds[4].format(self.vehiclesID["veh1"]), (newVel - self.realVehleadVelocity))
            self.matlabEng.eval(cmdExec, nargout=0)

    def startNextStep(self, action):  #(self, initAction, prevAction, action):
        """
        To run the code continuously till the end of SUMO traffic simulation
        :param action: New Acceleration value from RL
        :return: After the simulation, give us the Sensor Values
        """
        if action:
            self.Action(action) #(initAction, prevAction, action)
        # if action:
        #     actions = action/Constants.STEPSPERACTION

        # aggregatedStates = pd.DataFrame([], columns=["prevVehAcc", "hw", "hwdelta", "longSlip_FL", "longSlip_FR",
        #                                              "longSlip_RL", "longSlip_RR", "frictionMu_FL", "frictionMu_FR",
        #                                              "frictionMu_RL", "frictionMu_RR"
        #                                              ])
        aggregatedStates = pd.DataFrame([], columns=["egoVel", "delta_D", "delta_V", "longSlip_FL", "longSlip_FR",
                                                     "longSlip_RL", "longSlip_RR", "frictionMu_FL", "frictionMu_FR",
                                                     "frictionMu_RL", "frictionMu_RR"])
        aggregatedRewards = pd.DataFrame([], columns=["Rewards"])
        aggregatedInfo = []
        for curstep in range(Constants.STEPSPERACTION):
            # self.Action(actions, [True if curstep == 0 else False][0])
            self.simulationTime += self.timeStep
            self.leadVehicleAcceleration(self.simulationTime)
            _, sim_status = self.nextStep()
            vehicleSensorData = {i: {"RADAR": [], "Camera": []} for i in self.simulationVehIDs}
            relativeDist, relVelocity, leadVehID = 2, 0, self.VUT
            if sim_status:
                self.done = False
                # print("Updated Scenario File. Starting Simulink Execution")
                cmdExec = ""
                for idx, val in enumerate(self.simulationVehIDs, 1):
                    if self.vehDetectionInfo[val]["pos"][0] + 1.5375 >= self.fricChangePos[0] and \
                            not self.fricChangeInd[val][0][0]:
                        # front wheel change
                        self.frictionChange(2, val, self.fricChange[0])
                        self.fricChangeInd[val][0][0] = 1
                    elif self.vehDetectionInfo[val]["pos"][0] - 1.5375 >= self.fricChangePos[0] and \
                            not self.fricChangeInd[val][0][1]:
                        # rear wheel change
                        self.frictionChange(3, val, self.fricChange[0])
                        self.fricChangeInd[val][0][1] = 1
                    if self.vehDetectionInfo[val]["pos"][0] + 1.5375 >= self.fricChangePos[1] and \
                            not self.fricChangeInd[val][1][0]:
                        # front wheel revert
                        self.frictionChange(2, val, self.fricChange[1])
                        self.fricChangeInd[val][1][0] = 1
                    elif self.vehDetectionInfo[val]["pos"][0] - 1.5375 >= self.fricChangePos[1] and \
                            not self.fricChangeInd[val][1][1]:
                        # rear wheel revert
                        self.frictionChange(3, val, self.fricChange[1])
                        self.fricChangeInd[val][1][1] = 1
                    '''
                    if self.vehDetectionInfo[val]["speedChange"]:
                        cmdExec = cmdExec + \
                                  "set_param('{}', 'Value', '{}');".format(self.blkCmds[0].format(idx),
                                                                           str(max(
                                                                               self.vehDetectionInfo[val][
                                                                                   "speed"] + 0.0,
                                                                               min(Constants.VEL_BOUND))))
                    '''
                    if self.vehDetectionInfo[val]["laneChange"]:
                        cmdExec = cmdExec + \
                                  "set_param('{}', 'Value', '{}');".format(self.blkCmds[1].format(idx),
                                                                           str(self.vehDetectionInfo[val]["pos"][1]))
                # self.matlabEng.eval(cmdExec, nargout=0)

                executionCommand = cmdExec + Constants.CMDSETUP.format(
                    Constants.SIMULINKMODELS[0], Constants.SIMCMD, Constants.MATLABCOMMANDS[2]) + ";"
                executionCommand = executionCommand + \
                                   Constants.CMDSETUP.format(
                                       Constants.SIMULINKMODELS[1], Constants.SIMCMD, Constants.MATLABCOMMANDS[2]) + ";"
                self.matlabEng.eval(executionCommand, nargout=0)
                # print("Simulink Model Execution Completed")
                # get the current Sensor values from ObjectDetection Model
                for idx, val in enumerate(self.simulationVehIDs, 1):
                    if val in self.radarVehicles and val not in self.arrivedVehicles:
                        cmdToRun = [[self.simuGetData[val]["rto"][6] + Constants.INPUTPORT_SHORT],
                                    [self.simuGetData[val]["rto"][7] + Constants.INPUTPORT_SHORT]]
                        combinedOutput = self.matlabEng.eval("getRunTimeData{}".format(tuple(cmdToRun)), nargout=1)
                        radVal = combinedOutput[self.simuGetData[val]["rto"][6]]
                        camVal = combinedOutput[self.simuGetData[val]["rto"][7]]
                        # selecting only relative distance, velocity and detection id from sensors data
                        for i in range(self.vdObject.radarMaxDet):
                            vehicleSensorData[val]["RADAR"].append([radVal[0 + i * 3][0], radVal[1 + i * 3][0],
                                                                    radVal[2 + i * 3][0] if int(
                                                                        radVal[2 + i * 3][0]) not
                                                                                            in self.arrivedVehicles
                                                                    else -1])
                        for i in range(self.vdObject.cameraMaxDet):
                            vehicleSensorData[val]["Camera"].append([camVal[0 + i * 3][0], camVal[1 + i * 3][0],
                                                                     camVal[2 + i * 3][0] if int(
                                                                         camVal[2 + i * 3][0]) not
                                                                                             in self.arrivedVehicles
                                                                     else -1])
                    else:
                        vehicleSensorData[val]["RADAR"].append([0, 0, 0])
                        vehicleSensorData[val]["Camera"].append([0, 0, 0])

                # collect the sensor data from each vehicle and calculate suitable acceleration through ACC

                data = vehicleSensorData[self.VUT]["RADAR"]
                data.extend(vehicleSensorData[self.VUT]["Camera"])
                # print(data)
                finalSensordf = pd.DataFrame(data)
                falseIndexes = finalSensordf[(finalSensordf[2] == 0) | (finalSensordf[2] == -1)].index
                finalSensordf.drop(falseIndexes, inplace=True)
                # finding the min relative distance between the vehicles and choosing it as a target vehicle for ACC
                '''RIL code should come here'''
                if finalSensordf.shape[0]:
                    self.vehDetectionInfo[Constants.VIR_VEH]["active"] = False
                    finalData = list(finalSensordf.iloc[np.argmin(finalSensordf[0]), :])
                    relativeDist, relVelocity, leadVehID = finalData[0], finalData[1], \
                                                           [key for key, value in self.vehiclesID.items()
                                                            if value == finalData[2]][0]
                    self.leadVelocity = self.vehDetectionInfo[leadVehID]["initialSpeed"]
                    self.updateVDData(leadVehID)
                else:
                    # when no vehicles in range. Velocity control of ACC is activated to reach vehicle desired velocity
                    # for RL convenience introducing virtual vehicle with velocity of ego vehicle desired velocity
                    # with initial headway of 2 seconds. After creation update position every 0.1 seconds
                    curVirtualVehSpeed = np.array(self.vehDetectionInfo[Constants.VIR_VEH]["speedVector"])
                    self.leadVelocity = curVirtualVehSpeed[0]
                    self.updateVDData(self.VUT)
                    if not self.vehDetectionInfo[Constants.VIR_VEH]["active"]:
                        self.vehDetectionInfo[Constants.VIR_VEH]["active"] = True
                        self.vehDetectionInfo[Constants.VIR_VEH]["pos"] = self.vehDetectionInfo[self.VUT]["pos"] + \
                                                                          (2 * curVirtualVehSpeed)

                    else:
                        self.vehDetectionInfo[Constants.VIR_VEH]["pos"] = np.array(
                            self.vehDetectionInfo[Constants.VIR_VEH]["pos"]) + \
                                                                          np.array(self.timeStep * curVirtualVehSpeed)

                    # calculating ds, dv for virtual vehicle similar to the RADAR calculation in Simulink model
                    direcVector = np.array([math.cos(math.radians(0)), math.sin(math.radians(0)), 0])
                    relativeDist = (np.array(self.vehDetectionInfo[Constants.VIR_VEH]["pos"]) * direcVector - \
                                    np.array(self.vehDetectionInfo[self.VUT]["pos"]) * direcVector)[0]
                    relVelocity = (np.array(self.vehDetectionInfo[Constants.VIR_VEH]["speedVector"]) * direcVector - \
                                   np.array(self.vehDetectionInfo[self.VUT]["speedVector"]) * direcVector)[0]
                    leadVehID = self.VUT
                    vehicleSensorData["virtual"] = [relativeDist, relVelocity, leadVehID]
                self.sensorValues.append(vehicleSensorData)
                # update the Simulink values to SUMO. Only position of the vehicles are changing now. we can add more.
                try:
                    for i in self.simuGetData:
                        if i not in self.arrivedVehicles:
                            cmdToRun = [[self.simuGetData[i]["rto"][5] + Constants.INPUTPORT_SHORT],
                                        [self.simuGetData[i]["rto"][0] + Constants.INPUTPORT_SHORT],
                                        [self.simuGetData[i]["rto"][1] + Constants.INPUTPORT_SHORT]]
                            combinedOutput = self.matlabEng.eval("getRunTimeData{}".format(tuple(cmdToRun)), nargout=1)
                            speed = combinedOutput[self.simuGetData[i]["rto"][5]]
                            if speed > 0:
                                self.vehDetectionInfo[i]["speed"] = speed
                                curVehAngle = traci.vehicle.getAngle(i)
                                self.vehDetectionInfo[i]["speedVector"] = np.array(
                                    [round(sin(math.radians(curVehAngle)), 2),
                                     round(cos(math.radians(curVehAngle)), 2),
                                     0]) * \
                                                                          speed
                                posx = combinedOutput[self.simuGetData[i]["rto"][0]] - \
                                       self.vehDetectionInfo[i]["vdInitialPosition"]
                                posy = combinedOutput[self.simuGetData[i]["rto"][1]]
                                traci.vehicle.setSpeed(i, speed)
                                traci.vehicle.setSpeedMode(i, 0)
                                traci.vehicle.moveToXY(i, traci.vehicle.getLaneID(i), traci.vehicle.getLaneIndex(i),
                                                       x=self.sumoInfo.sumoNetworkValues["sumoOffset"][0] + round(
                                                           float(posx),
                                                           2),
                                                       y=self.sumoInfo.sumoNetworkValues["sumoOffset"][1] + posy)
                            else:
                                self.vehDetectionInfo[i]["speed"] = 0
                                traci.vehicle.setSpeed(i, 0)
                                traci.vehicle.setSpeedMode(i, 0)
                    traci.simulationStep()
                except Exception as e:
                    print("Road is over for vehicle {}. Move Vehicle request locations: {},{}".
                          format(i, self.sumoInfo.sumoNetworkValues["sumoOffset"][0] + round(float(posx), 2),
                                 self.sumoInfo.sumoNetworkValues["sumoOffset"][1] + posy))

            else:
                self.done = True
                # calculate states based on last know position
                direcVector = np.array([math.cos(math.radians(0)), math.sin(math.radians(0)), 0])
                relativeDist = (np.array(self.vehDetectionInfo[Constants.VIR_VEH]["pos"]) * direcVector - \
                                np.array(self.vehDetectionInfo[self.VUT]["pos"]) * direcVector)[0]
                relVelocity = (np.array(self.vehDetectionInfo[Constants.VIR_VEH]["speedVector"]) * direcVector - \
                               np.array(self.vehDetectionInfo[self.VUT]["speedVector"]) * direcVector)[0]
                leadVehID = self.VUT
                self.updateVDData(leadVehID)

            self.hw = round(relativeDist / max(self.egoVelocity, 0.1), 4)
            self.delta_distance = relativeDist #lead - ego
            self.hwdelta = round(self.hw - self.preHW, 4)
            self.preHW = self.hw
            self.jerk = round((self.egoVehAcc - self.egoPrevAcc) / self.timeStep, 4)
            self.egoPrevAcc = round(self.egoVehAcc, 4)
            self.ttc = [relativeDist / abs(relVelocity)
                        if relVelocity < 0 else 10000][0]
            self.delta_velocity = relVelocity #lead - ego
            self.leadvehspeed = self.egoVelocity + relVelocity
            # self.boundingStates()
            '''states = np.array([self.prevVehAcc, self.hw, self.hwdelta,
                               self.longSlip_FL, self.longSlip_FR, self.longSlip_RL, self.longSlip_RR,
                               self.frictionMu_FL, self.frictionMu_FR, self.frictionMu_RL, self.frictionMu_RR,
                               ])'''
            states = np.array([round(self.egoVelocity, 4), round(self.delta_distance, 4), 
                               round(self.delta_velocity, 4), 
                               round(self.longSlip_FL, 4), round(self.longSlip_FR, 4), 
                               round(self.longSlip_RL, 4), round(self.longSlip_RR, 4),
                               round(self.frictionMu_FL, 4), round(self.frictionMu_FR, 4), 
                               round(self.frictionMu_RL, 4), round(self.frictionMu_RR, 4)])

            '''reward_for_action, rwSplitUp = self.calculate_reward(self.hw, self.jerk,
                                                                 self.longSlip_FL, self.longSlip_FR,
                                                                 self.longSlip_RL, self.longSlip_RR,
                                                                 self.egoVelocity)'''
            
            w = [1, 1, 0.5, 0.5]  #safe; speed; stable; comfort
            reward_for_action, rwSplitUp = self.rewardCal(self.jerk, self.delta_distance, 
                                                          self.leadvehspeed, self.delta_velocity,
                                                          self.longSlip_FL, self.longSlip_FR,
                                                          self.longSlip_RL, self.longSlip_RR,
                                                          w)
            
            
            
            '''info = [self.egoVelocity, self.hw, self.hwdelta, self.egoVehAcc, self.longSlip, self.jerk, self.ttc,
                    rwSplitUp]'''
    
            info = [self.egoVelocity, self.hw, self.prevVehAcc, self.egoVehAcc, self.longSlip, self.jerk, self.leadvehspeed,
                    rwSplitUp]
            
            if self.initialStep:
                self.initialStep = False
            aggregatedStates.loc[len(aggregatedStates), :] = states
            aggregatedRewards.loc[len(aggregatedRewards), :] = reward_for_action
            aggregatedInfo.append(info)

            if self.done:
                break
        # aggregating multiple states into one
        # moving average

        # weighted average
        states = [
            round(np.sum(aggregatedStates.iloc[:, 0] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 1] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 2] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 3] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 4] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 5] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 6] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 7] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 8] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 9] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4),
            round(np.sum(aggregatedStates.iloc[:, 10] * Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]) /
                  np.sum(Constants.WGHTSFORSTATEAGG[:len(aggregatedStates)]), 4)
        ]
        # aggregating multiple rewards into one
        # average
        reward_for_action = round(np.mean(aggregatedRewards)[0], 4)
        # weighted mean states, mean rewards, combined Observations,

        # done signal # break the loop if done signal is received in between
        return (states, reward_for_action, self.done, aggregatedInfo)

    def Action(self, acceleration): #(self, initial, prevAcc, acceleration):
        # moving average window 5
        # self.actions.loc[len(self.actions), "acceleration"] = acceleration[0]
        # if self.actions.shape[0] >= 3:
        #     accMA = self.actions.rolling(window=3).mean()
        #     chosenAcc = list(accMA.values)[-1][0]
        # else:
        chosenAcc = acceleration[0]
        cmdExec = ""
        cmdExec = cmdExec + "set_param('{}', 'Value', '{}');". \
            format(self.blkCmds[4].format(self.vehiclesID[self.VUT]),
                   chosenAcc)  # * self.timeStep * Constants.STEPSPERACTION)
        self.matlabEng.eval(cmdExec, nargout=0)

    def boundingStates(self):
        # self.prevVehAcc = max(min(self.prevVehAcc, action_space.high[0]), action_space.low[0])
        self.egoVelocityNorm = self.egoVelocity
        # self.egoVelocityNorm = self.egoVelocity / (max(Constants.VEL_BOUND) - min(Constants.VEL_BOUND))
        self.hw = max(min(self.hw, max(Constants.HW_BOUND)), min(Constants.HW_BOUND))
        self.hwdelta = max(min(self.hwdelta, max(Constants.HWDELTA_BOUND)), min(Constants.HWDELTA_BOUND))

    def updateVDData(self, leadVehID):
        cmdToRun = [[self.simuGetData[self.VUT]["rto"][5] + Constants.INPUTPORT_SHORT],
                    [self.simuGetData[self.VUT]["rto"][2] + Constants.INPUTPORT_SHORT],
                    [self.simuGetData[self.VUT]["rto"][3] + Constants.INPUTPORT_SHORT],
                    [self.simuGetData[self.VUT]["rto"][9] + Constants.INPUTPORT_SHORT],
                    [self.simuGetData[self.VUT]["rto"][8] + Constants.INPUTPORT_SHORT],
                    [self.simuGetData[leadVehID]["rto"][5] + Constants.INPUTPORT_SHORT]]
        combinedOutput = self.matlabEng.eval("getRunTimeData{}".format(tuple(cmdToRun)), nargout=1)
        # ego vehicle current velocity
        self.egoVelocity = round(combinedOutput[self.simuGetData[self.VUT]["rto"][5]], 2)
        cmdToRun = self.simuGetData["veh1"]["rto"][5] + Constants.INPUTPORT
        self.realVehleadVelocity = round(self.matlabEng.eval(cmdToRun, nargout=1), 4)
        # ego vehicle current longitudinal slip
        self.longSlip = combinedOutput[self.simuGetData[self.VUT]["rto"][2]]
        # ego vehicle friction coefficient
        self.fullFriction = combinedOutput[self.simuGetData[self.VUT]["rto"][3]]
        self.frictionMu_FL = self.fullFriction[0][0]
        self.frictionMu_FR = self.fullFriction[1][0]
        self.frictionMu_RL = self.fullFriction[2][0]
        self.frictionMu_RR = self.fullFriction[3][0]
        # Ego Vehicle Acceleration
        self.egoVehAcc = combinedOutput[self.simuGetData[self.VUT]["rto"][9]]
        # All Slip
        self.fullSlip = combinedOutput[self.simuGetData[self.VUT]["rto"][8]]
        self.longSlip_FL = self.fullSlip[0][0]
        self.longSlip_FR = self.fullSlip[0][1]
        self.longSlip_RL = self.fullSlip[0][2]
        self.longSlip_RR = self.fullSlip[0][3]
        # Lead Vehicle Acceleration - Virtual vehicles have "0" acceleration
        if self.traciInitial:
            cmdToRun = self.simuGetData[leadVehID]["rto"][9] + Constants.INPUTPORT
            self.prevVehAcc = round(self.matlabEng.eval(cmdToRun, nargout=1), 4)
        else:
            if not self.vehDetectionInfo[Constants.VIR_VEH]["active"]:
                cmdToRun = self.simuGetData[leadVehID]["rto"][9] + Constants.INPUTPORT
                self.prevVehAcc = round(self.matlabEng.eval(cmdToRun, nargout=1), 4)
            else:
                self.prevVehAcc = 0

    def rewardSplit(self, hw, jerk, longSlip_FL, longSlip_FR, longSlip_RL, longSlip_RR):
        if 1.25 <= hw <= 1.35:
            x1 = True
        else:
            x1 = False
        if 0 <= abs(jerk) <= 0.9:
            x2 = True
        else:
            x2 = False
        x3 = []
        if 0 <= abs(longSlip_FL) <= 0.2:
            x3.append(True)
        else:
            x3.append(False)
        if 0 <= abs(longSlip_FR) <= 0.2:
            x3.append(True)
        else:
            x3.append(False)
        if 0 <= abs(longSlip_RL) <= 0.2:
            x3.append(True)
        else:
            x3.append(False)
        if 0 <= abs(longSlip_RR) <= 0.2:
            x3.append(True)
        else:
            x3.append(False)
        x3 = [True if len(np.where(np.array(x3))[0]) == 4 else False][0]
        weightages = self.splitUp[(x1, x2, x3)]
        return (weightages)

    def rewardSplitNoJerk(self, hw, longSlip_FL, longSlip_FR, longSlip_RL, longSlip_RR):
        if 1.25 <= hw <= 1.35:
            x1 = True
        else:
            x1 = False

        x3 = []
        if 0 <= abs(longSlip_FL) <= 0.2:
            x3.append(True)
        else:
            x3.append(False)
        if 0 <= abs(longSlip_FR) <= 0.2:
            x3.append(True)
        else:
            x3.append(False)
        if 0 <= abs(longSlip_RL) <= 0.2:
            x3.append(True)
        else:
            x3.append(False)
        if 0 <= abs(longSlip_RR) <= 0.2:
            x3.append(True)
        else:
            x3.append(False)
        x3 = [True if len(np.where(np.array(x3))[0]) == 4 else False][0]
        weightages = self.splitUpNoJerk[(x1, x3)]
        return (weightages)

    def calculate_reward(self, hw, jerk, longSlip_FL, longSlip_FR, longSlip_RL, longSlip_RR, velocity):
        if self.ttc > 4:
            w_comfort = 1
            weights = self.rewardSplit(hw, jerk, longSlip_FL, longSlip_FR, longSlip_RL, longSlip_RR)
            hw_reward = weights[0] * self.headwayReward(hw)

            cmfrt_reward = weights[1] * (w_comfort * self.comfortReward(abs(jerk)))

            stab_reward = weights[2] * self.stabilityReward(abs(longSlip_FL),
                                                            abs(longSlip_FR),
                                                            abs(longSlip_RL),
                                                            abs(longSlip_RR))

            tot_rew = hw_reward + stab_reward + cmfrt_reward + self.desiredVel(velocity)
            return (round(tot_rew, 4), [weights, hw_reward, cmfrt_reward, stab_reward])
        else:
            w_comfort = 0
            weights = self.rewardSplitNoJerk(hw, longSlip_FL, longSlip_FR, longSlip_RL, longSlip_RR)
            hw_reward = weights[0] * self.headwayReward(hw)
            stab_reward = weights[1] * self.stabilityReward(abs(longSlip_FL),
                                                            abs(longSlip_FR),
                                                            abs(longSlip_RL),
                                                            abs(longSlip_RR))
            tot_rew = hw_reward + stab_reward + self.desiredVel(velocity)
            return (round(tot_rew, 4), [weights, hw_reward, stab_reward])

    def rewardCal(self, jerk, delta_distance, leadvehspeed, delta_velocity,
                  longSlip_FL, longSlip_FR, longSlip_RL, longSlip_RR, w):
        velocity = leadvehspeed - delta_velocity
        safe_rewd = w[0] * (self.safeReward(delta_distance, leadvehspeed) + 0.5) * 2 #[-1, 1]
        speed_rewd = w[1] * ((math.exp(-abs(delta_velocity))-1) +0.5) * 2 #[-1, 1]
        stable_rewd = w[2] * (self.stableReward(abs(longSlip_FL),
                                               abs(longSlip_FR),
                                               abs(longSlip_RL),
                                               abs(longSlip_RR)) + 0.5) * 2  #[-1, 1]
        comfort_rewd = w[3] * (self.comfortReward(abs(jerk))+0.5) * 2 #(-1, 1]
        tot_reward = safe_rewd + speed_rewd + stable_rewd +  comfort_rewd + self.desiredVel(velocity)
        
        tot_reward_mpd, safe_rewd_mpd, speed_rewd_mpd, stable_rewd_mpd, comfort_rewd_mpd = \
                                    self.lmap(tot_reward, safe_rewd, speed_rewd, stable_rewd,  comfort_rewd,
                                                  [(-1*w[0]) + (-1*w[1]) + (-1*w[2]) + (-1*w[3]), (1*w[0]) + (1*w[1]) + (1*w[2]) + (1*w[3])],
                                                  [-1, 1]) 
        
        return  (tot_reward_mpd, [w, safe_rewd_mpd, speed_rewd_mpd, stable_rewd_mpd, comfort_rewd_mpd])        
        
     
    
    def desiredVel(self, velocity):
        if velocity <= 0.2:
            self.zeroVelCounter += 1
            if self.zeroVelCounter >= 50:
                self.done = True
                #print("ending episode due to zero velocity")
            return (0)
        else:
            self.zeroVelCounter = 0
            return (0)

    def headwayReward(self, hw):
        if hw <= 0.5:
            self.done = True
            return -1
        else:
            '''
            mu, sigma = 0.715, 0.15
            hw_rwValue = 0.7600 * (
                    np.exp(-(np.log(hw) - mu) ** 2 / (2 * sigma ** 2)) / (hw * sigma * np.sqrt(2 * np.pi)))
            hw_rwValue = 2 * (hw_rwValue - 0.5)
            '''
            mu, sigma = 0.285, 0.15
            hw_rwValue = 0.4944 * (
                    np.exp(-(np.log(hw) - mu) ** 2 / (2 * sigma ** 2)) / (hw * sigma * np.sqrt(2 * np.pi)))
            hw_rwValue = 2 * (hw_rwValue - 0.5)
            hw_rwValue = max(min(hw_rwValue, 1), -1)
        return (round(hw_rwValue, 4))

    def stabilityReward(self, longSlip_FL, longSlip_FR, longSlip_RL, longSlip_RR):
        stab_rwValue = []
        stab_rwValue.append(2.0098 * math.tanh(-3 * longSlip_FL) + 1)
        stab_rwValue.append(2.0098 * math.tanh(-3 * longSlip_FR) + 1)
        stab_rwValue.append(2.0098 * math.tanh(-3 * longSlip_RL) + 1)
        stab_rwValue.append(2.0098 * math.tanh(-3 * longSlip_RR) + 1)
        minStab_rwValue = min(stab_rwValue)
        return (round(minStab_rwValue, 4))
    
    def comfortReward(self, jerk):
        cmfrt_rwValue = - mem.smf(np.array([jerk]), 0.6, 2)
        return (round(cmfrt_rwValue[0], 4))

    # def comfortReward(self, jerk):
    #     popt = np.array([-0.17385645, 1.83301304, -8.41317618, 21.44826674,
    #                      -30.8967516, 23.06809265, -7.90855301, 0.70841497,
    #                      0.99938499])

    #     def fn(x, a, b, c, d, e, f, g, h, i):
    #         out = a * (x ** 8) + b * (x ** 7) + c * (x ** 6) + d * (x ** 5) + e * (x ** 4) + f * (x ** 3) + g * (
    #                     x ** 2) + h * (x ** 1) + i * (x ** 0)
    #         return (out)
    #     if jerk >= 2:
    #         out = -1
    #     else:
    #         out = round(fn(jerk, popt[0], popt[1], popt[2], popt[3], popt[4], popt[5], popt[6], popt[7], popt[8]), 2)

    #     cmfrt_rwValue = max(min(out, 1), -1)
    #     # cmfrt_rwValue = 1 - mem.smf(np.array([jerk]), 0.6, 2)
    #     # cmfrt_rwValue = 2 * (cmfrt_rwValue - 0.5)
    #     return (round(cmfrt_rwValue, 4))
    
    def safeReward(self, delta_distance, leadvehspeed):
        safe_distance = 3*leadvehspeed #empirical value
        
        if delta_distance < 0.2*safe_distance:
            safe_rewd = -1
            self.done = True
            return  safe_rewd 
        
        else:
            if 1.8*safe_distance <= delta_distance :
                safe_rewd = -1
                return safe_rewd 
            
            elif  0.2*safe_distance <= delta_distance and delta_distance < 0.5*safe_distance:
                A = 0.8 / ((0.3 * safe_distance) ** 2)
                safe_rewd = A * (delta_distance - 0.2 * safe_distance) ** 2 - 1
                return safe_rewd 
                
            elif 0.5*safe_distance <= delta_distance and delta_distance < 1.5*safe_distance:
                A = -0.2 / ((0.5 * safe_distance) ** 2)
                safe_rewd = A * (delta_distance - safe_distance)**2
                return safe_rewd 
            
            elif 1.5*safe_distance <= delta_distance and delta_distance < 1.8*safe_distance:
                A = 0.8 / ((0.3 * safe_distance) ** 2)
                safe_rewd = A * (delta_distance - 1.8 * safe_distance)**2 - 1
                return safe_rewd     
    
    def stableReward(self, longSlip_FL, longSlip_FR, longSlip_RL, longSlip_RR):
        stab_rwValue = []
        stab_rwValue.append(1.0049 * math.tanh(-3 * longSlip_FL))
        stab_rwValue.append(1.0049 * math.tanh(-3 * longSlip_FR))
        stab_rwValue.append(1.0049 * math.tanh(-3 * longSlip_RL))
        stab_rwValue.append(1.0049 * math.tanh(-3 * longSlip_RR))
        minStab_rwValue = min(stab_rwValue)
        return (round(minStab_rwValue, 4))
    
    def lmap(self, tot_reward, safe_rewd, speed_rewd, stable_rewd,  comfort_rewd, x, y):
        """Linear map of input value with range x to desired range y."""
        k =  (y[1] - y[0]) / (x[1] - x[0])
        tot_reward_mpd = y[0] + (tot_reward - x[0]) * k
        safe_rewd_mpd = y[0] + (safe_rewd - x[0]) * k - 0.75 * y[0] + 0.75 * k * x[0]
        speed_rewd_mpd = y[0] + (speed_rewd - x[0]) * k - 0.75 * y[0] + 0.75 * k * x[0]
        stable_rewd_mpd = y[0] + (stable_rewd - x[0]) * k - 0.75 * y[0] + 0.75 * k * x[0]
        comfort_rewd_mpd = y[0] + (comfort_rewd - x[0]) * k - 0.75 * y[0] + 0.75 * k * x[0]
        
        return round(tot_reward_mpd,4), round(safe_rewd_mpd,4), \
            round(speed_rewd_mpd,4), round(stable_rewd_mpd,4), round(comfort_rewd_mpd,4)    
    
    
       
    def stopSimulation(self, mode):
        try:
            self.vdObject.deleteDetectionsModels(list(self.radarVehicles.values()))
            self.matlabEng.eval(
                Constants.CMDSETUP.format(Constants.SIMULINKMODELS[1], Constants.SIMCMD, Constants.MATLABCOMMANDS[-1]),
                nargout=0)
            if mode == "stop":
                # deleteModels after the simulation
                self.vdObject.deleteVDModels(self.vehCnt)
                self.matlabEng.exit()
            traci.close()
        except:
            print("SUMO TraCI might not be active yet")
            pass
