import torch as T
import torch.autograd
import torch.optim as optim
import sys
import numpy as np
import pandas as pd
from CoMoVe_RL_MultipleSteps_acc_CutOut import SimulationEnvironment
import time
from tqdm import tqdm


# %%
class Memory:
    def __init__(self, max_size, batch_size, num_states, num_actions, savedBuffers={}):
        # Number of "experiences" to store at max
        self.buffer_capacity = max_size
        # Num of tuples to train on.
        self.batch_size = batch_size

        # Its tells us num of times record() was called.
        self.buffer_counter = 0

        # Instead of list of tuples as the exp.replay concept go
        # We use different np.arrays for each tuple element

        if not len(savedBuffers):
            self.state_buffer = np.zeros((self.buffer_capacity, num_states))
            self.action_buffer = np.zeros((self.buffer_capacity, num_actions))
            self.reward_buffer = np.zeros((self.buffer_capacity, 1))
            self.next_state_buffer = np.zeros((self.buffer_capacity, num_states))
            self.done_state_buffer = np.zeros((self.buffer_capacity, 1))
        else:
            self.state_buffer = savedBuffers["states"]
            self.action_buffer = savedBuffers["actions"]
            self.reward_buffer = savedBuffers["rewards"]
            self.next_state_buffer = savedBuffers["nextStates"]
            self.done_state_buffer = savedBuffers["done"]

    def push(self, state, action, reward, next_state, done):
        index = self.buffer_counter % self.buffer_capacity

        self.state_buffer[index] = state
        self.action_buffer[index] = action
        self.reward_buffer[index] = reward
        self.next_state_buffer[index] = next_state
        self.done_state_buffer[index] = done

        self.buffer_counter += 1

    def sample(self):
        record_range = min(self.buffer_counter, self.buffer_capacity)
        # Randomly sample indices
        batch_indices = np.random.choice(record_range,
                                         [self.batch_size if record_range >= self.batch_size else record_range][0],
                                         replace=False)

        # Convert to tensors
        state_batch = self.state_buffer[batch_indices]
        action_batch = self.action_buffer[batch_indices]
        reward_batch = self.reward_buffer[batch_indices]
        next_state_batch = self.next_state_buffer[batch_indices]
        done_state_batch = self.done_state_buffer[batch_indices]
        return (state_batch, action_batch, reward_batch, next_state_batch, done_state_batch)

    def __len__(self):
        return self.buffer_counter


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.autograd
from torch.autograd import Variable
import os


class Critic(nn.Module):
    def __init__(self, input_size, hidden_size, output_size, state_size):
        super(Critic, self).__init__()
        self.critic1_stlinear1 = nn.Linear(state_size, 16)
        self.critic1_stlinear2 = nn.Linear(16, 32)
        self.critic1_aclinear1 = nn.Linear(output_size, 32)
        self.critic1_linear1 = nn.Linear(64, hidden_size)
        self.critic1_linear2 = nn.Linear(hidden_size, hidden_size)
        self.critic1_linear3 = nn.Linear(hidden_size, output_size)

        self.critic2_stlinear1 = nn.Linear(state_size, 16)
        self.critic2_stlinear2 = nn.Linear(16, 32)
        self.critic2_aclinear1 = nn.Linear(output_size, 32)
        self.critic2_linear1 = nn.Linear(64, hidden_size)
        self.critic2_linear2 = nn.Linear(hidden_size, hidden_size)
        self.critic2_linear3 = nn.Linear(hidden_size, output_size)
        self.checkpoint_dir = r"pyt_Models/TD3_cutOut_Scenario/trial_1"

    def forward(self, state, action):
        """
        Params state and actions are torch tensors
        """
        state_x = F.relu(self.critic1_stlinear1(state))
        state_x = F.relu(self.critic1_stlinear2(state_x))
        action_x = F.relu(self.critic1_aclinear1(action))
        q1 = torch.cat([state_x, action_x], 1)
        q1 = F.relu(self.critic1_linear1(q1))
        q1 = F.relu(self.critic1_linear2(q1))
        q1 = self.critic1_linear3(q1)

        state_x = F.relu(self.critic2_stlinear1(state))
        state_x = F.relu(self.critic2_stlinear2(state_x))
        action_x = F.relu(self.critic2_aclinear1(action))
        q2 = torch.cat([state_x, action_x], 1)
        q2 = F.relu(self.critic2_linear1(q2))
        q2 = F.relu(self.critic2_linear2(q2))
        q2 = self.critic2_linear3(q2)
        return q1, q2

    def critic_1(self, state, action):
        state_x = F.relu(self.critic1_stlinear1(state))
        state_x = F.relu(self.critic1_stlinear2(state_x))
        action_x = F.relu(self.critic1_aclinear1(action))
        q1 = torch.cat([state_x, action_x], 1)
        q1 = F.relu(self.critic1_linear1(q1))
        q1 = F.relu(self.critic1_linear2(q1))
        q1 = self.critic1_linear3(q1)
        return q1

    def save_checkpoint(self, name):
        # print('... saving checkpoint ...')
        checkpoint_file = os.path.join(self.checkpoint_dir, name + '_TD3')
        T.save(self.state_dict(), checkpoint_file)

    def load_checkpoint(self, checkpoint_file):
        print('... loading checkpoint ...')
        self.load_state_dict(T.load(checkpoint_file))

    def save_best(self, name):
        print('... saving best checkpoint ...')
        checkpoint_file = os.path.join(self.checkpoint_dir, name + '_TD3_best')
        T.save(self.state_dict(), checkpoint_file)


class Actor(nn.Module):
    def __init__(self, input_size, hidden_size, output_size, learning_rate=3e-4):
        super(Actor, self).__init__()
        self.linear1 = nn.Linear(input_size, hidden_size)
        self.linear2 = nn.Linear(hidden_size, hidden_size)
        self.linear3 = nn.Linear(hidden_size, output_size)
        self.checkpoint_dir = r"pyt_Models/TD3_cutOut_Scenario/trial_1"

    def forward(self, state):
        """
        Param state is a torch tensor
        """
        x = F.relu(self.linear1(state))
        x = F.relu(self.linear2(x))
        x = torch.tanh(self.linear3(x))
        x = x * 2
        return x

    def save_checkpoint(self, name):
        # print('... saving checkpoint ...')
        checkpoint_file = os.path.join(self.checkpoint_dir, name + '_TD3')
        T.save(self.state_dict(), checkpoint_file)

    def load_checkpoint(self, checkpoint_file):
        print('... loading checkpoint ...')
        self.load_state_dict(T.load(checkpoint_file))

    def save_best(self, name):
        print('... saving best checkpoint ...')
        checkpoint_file = os.path.join(self.checkpoint_dir, name + '_TD3_best')
        T.save(self.state_dict(), checkpoint_file)


# %%
class TD3agent:
    def __init__(self, env, hidden_size=64, actor_learning_rate=1e-3, critic_learning_rate=1e-3, gamma=0.99,
                 tau=1e-3, max_memory_size=int(50000), batch_size=48, loadModel=False,
                 policy_noise=0.1,
                 noise_clip=0.5,
                 policy_freq=2):
        # Params
        self.num_states = env.observation_space.shape[0]
        self.num_actions = env.action_space.shape[0]
        self.gamma = gamma
        self.tau = tau
        self.policyNoise = policy_noise
        self.noiseClip = noise_clip
        self.policyFreq = policy_freq
        self.policyIterationCount = 0

        # Networks
        self.actor = Actor(self.num_states, hidden_size, self.num_actions)
        self.actor_target = Actor(self.num_states, hidden_size, self.num_actions)
        self.critic = Critic(self.num_states + self.num_actions, hidden_size, self.num_actions, self.num_states)
        self.critic_target = Critic(self.num_states + self.num_actions, hidden_size, self.num_actions, self.num_states)
        if loadModel:
            self.actor.load_checkpoint(
                "/home/krishdin/Documents/PhDFiles/Projects/CoMoVe/pythonFiles/ddpg_models/Jan14/CutOut_V2V_Try7/mainActor_1624957087320_200_ddpg")
            self.actor_target.load_checkpoint(
                "/home/krishdin/Documents/PhDFiles/Projects/CoMoVe/pythonFiles/ddpg_models/Jan14/CutOut_V2V_Try7/tgtActor_1624957087320_200_ddpg")
            self.critic.load_checkpoint(
                "/home/krishdin/Documents/PhDFiles/Projects/CoMoVe/pythonFiles/ddpg_models/Jan14/CutOut_V2V_Try7/mainCritic_1624957087320_200_ddpg")
            self.critic_target.load_checkpoint(
                "/home/krishdin/Documents/PhDFiles/Projects/CoMoVe/pythonFiles/ddpg_models/Jan14/CutOut_V2V_Try7/tgtCritic_1624957087320_200_ddpg")
        else:
            for target_param, param in zip(self.actor_target.parameters(), self.actor.parameters()):
                target_param.data.copy_(param.data)

            for target_param, param in zip(self.critic_target.parameters(), self.critic.parameters()):
                target_param.data.copy_(param.data)

        # Training
        self.memory = Memory(max_memory_size, batch_size, self.num_states, self.num_actions)
        self.critic_criterion = nn.MSELoss()
        self.actor_optimizer = optim.Adam(self.actor.parameters(), lr=actor_learning_rate)
        self.critic_optimizer = optim.Adam(self.critic.parameters(), lr=critic_learning_rate)  # , weight_decay=0.01)

    def get_action(self, state):
        state = Variable(torch.from_numpy(np.array(state)).float().unsqueeze(0))
        action = self.actor.forward(state)
        action = action.detach().numpy().flatten()
        return action

    def update(self):
        self.policyIterationCount += 1
        states, actions, rewards, next_states, done_ = self.memory.sample()
        states = torch.FloatTensor(states)
        actions = torch.FloatTensor(actions)
        rewards = torch.FloatTensor(rewards)
        next_states = torch.FloatTensor(next_states)
        doneState = torch.FloatTensor(done_)

        # we don't call backward or apply backpropagation for target networks, so we can use no_grad to speed up the
        # computation
        with torch.no_grad():
            # Select action according to policy and add clipped noise
            noise = (
                    torch.randn_like(actions) * self.policyNoise
            ).clamp(-self.noiseClip, self.noiseClip)

            next_action = (
                    self.actor_target.forward(next_states) + noise
            ).clamp(env.action_space.low[0], env.action_space.high[0])

            # Compute the target Q value
            target_Q1, target_Q2 = self.critic_target.forward(next_states, next_action)
            target_Q = torch.min(target_Q1, target_Q2)
            target_Q = rewards + self.gamma * target_Q * (1 - doneState)

        # Get current Q estimates
        current_Q1, current_Q2 = self.critic.forward(states, actions)

        # Compute critic loss
        critic_loss = F.mse_loss(current_Q1, target_Q) + F.mse_loss(current_Q2, target_Q)

        # Optimize the critic
        self.critic_optimizer.zero_grad()
        critic_loss.backward()
        self.critic_optimizer.step()

        # Delayed policy updates
        if self.policyIterationCount % self.policyFreq == 0:

            # Compute actor losse
            actor_loss = -self.critic.critic_1(states, self.actor(states)).mean()

            # Optimize the actor
            self.actor_optimizer.zero_grad()
            actor_loss.backward()
            self.actor_optimizer.step()

            # Update the frozen target models
            for target_param, param in zip(self.actor_target.parameters(), self.actor.parameters()):
                target_param.data.copy_(param.data * self.tau + target_param.data * (1.0 - self.tau))

            for target_param, param in zip(self.critic_target.parameters(), self.critic.parameters()):
                target_param.data.copy_(param.data * self.tau + target_param.data * (1.0 - self.tau))

    def clip_grad_norm_(self, module, max_grad_norm):
        nn.utils.clip_grad_norm_([p for g in module.param_groups for p in g["params"]], max_grad_norm)

    def save_models(self, name):
        self.actor.save_checkpoint("mainActor_" + name)
        self.actor_target.save_checkpoint("tgtActor_" + name)
        self.critic.save_checkpoint("mainCritic_" + name)
        self.critic_target.save_checkpoint("tgtCritic_" + name)
        checkpoint_file = os.path.join(self.actor.checkpoint_dir, name + '_actor_optimizer')
        T.save(self.actor_optimizer.state_dict(), checkpoint_file)
        checkpoint_file = os.path.join(self.actor.checkpoint_dir, name + '_critic_optimizer')
        T.save(self.critic_optimizer.state_dict(), checkpoint_file)

    def save_bestmodels(self, name):
        self.actor.save_best("mainActor_" + name)
        self.actor_target.save_best("tgtActor_" + name)
        self.critic.save_best("mainCritic_" + name)
        self.critic_target.save_best("tgtCritic_" + name)
        checkpoint_file = os.path.join(self.actor.checkpoint_dir, name + '_best_actor_optimizer')
        T.save(self.actor_optimizer.state_dict(), checkpoint_file)
        checkpoint_file = os.path.join(self.actor.checkpoint_dir, name + '_best_critic_optimizer')
        T.save(self.critic_optimizer.state_dict(), checkpoint_file)


# %%
pythonFilePath = r"/home/cy/Documents/PhD_Files/projects/CoMoVe_Framework/pythonFiles"
np.random.seed(0)
T.manual_seed(0)
env = SimulationEnvironment()
agent = TD3agent(env)
ep_rewards = []
epiStates = []
epiActions = []
epiInfo = []
epiRewards = []
SHOW_STATS_EVERY = 50
AGGREGATE_STATS_EVERY = 5
step = 0
num_episodes = 1500
num_steps = 150
best_score = 0
startEpi = 1
stepsCnt = 0
for episode in tqdm(range(1, num_episodes + 1)):
    try:
        state = env.reset()
    except Exception as e:
        print(f"{e}")
        raise ValueError("State Reset Error")
    episode_reward = []
    episode_actions = []
    episode_states = []
    episode_info = []
    initAction = True
    for step in range(num_steps):
        stepsCnt = stepsCnt + 1
        action = agent.get_action(state)
        action = np.clip(action, env.action_space.low[0], env.action_space.high[0])
        try:
            new_state, reward, done, info = env.startNextStep(action)
        except Exception as e:
            print(f"{e}")
            raise ValueError("Exception in start next step")
        agent.memory.push(state, action, reward, new_state, done)
        agent.update()
        state = new_state
        episode_reward.append(reward)
        episode_actions.append(action)
        episode_states.append(new_state)
        episode_info.append(info)

        if done:
            break

        if not step % SHOW_STATS_EVERY:
            sys.stdout.write("episode: {}, steps:{}, reward: {}\n".
                             format(episode, step, np.sum(episode_reward[-SHOW_STATS_EVERY:])))

    sys.stdout.write("episode: {}, steps:{}, reward: {}, average _reward: {} \n".
                     format(episode, step, np.sum(episode_reward), np.mean(episode_reward)))
    ep_rewards.append(sum(episode_reward))
    epiStates.append(episode_states)
    epiActions.append(episode_actions)
    epiInfo.append(episode_info)
    epiRewards.append(episode_reward)

    if not episode % AGGREGATE_STATS_EVERY or episode == 1:
        # if episode = 1:
        df = pd.DataFrame(epiStates)
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/Episode_{}-{}_States_normRew_{}.csv".format(startEpi, episode,
                                                                                     int(
                                                                                         time.time() * 1000)))
        df = pd.DataFrame(epiActions)
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/Episode_{}-{}_Actions_normRew_{}.csv".format(startEpi, episode,
                                                                                      int(
                                                                                          time.time() * 1000)))
        df = pd.DataFrame(epiInfo)
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/Episode_{}-{}_Info_normRew_{}.csv".format(startEpi, episode,
                                                                                   int(time.time() * 1000)))
        df = pd.DataFrame(epiRewards)
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/Episode_{}-{}_rewards_normRew_{}.csv".format(startEpi, episode,
                                                                                      int(
                                                                                          time.time() * 1000)))
        df = None
        startEpi = episode + 1
        epiStates = []
        epiActions = []
        epiInfo = []
        epiRewards = []
    agent.save_models("{}_{}".format(int(time.time() * 1000), episode))
    if step + 1 >= num_steps:
        env.matlabEng.eval("saveSDIData('{}', {})".format((f"{pythonFilePath}/logs/TD3_cutOut_Scenario/trial_1/"), episode),
                           nargout=0)
    if np.mean(
            episode_reward) > best_score:  # (np.sum(episode_reward) / (num_steps)) > best_score:  # and step==args.steps-1
        df = pd.DataFrame(episode_states)
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/bestEpisode-{}_States_normRew_{}.csv".format(episode,
                                                                                      int(
                                                                                          time.time() * 1000)))
        df = pd.DataFrame(episode_actions)
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/bestEpisode-{}_Actions_normRew_{}.csv".format(episode,
                                                                                       int(
                                                                                           time.time() * 1000)))
        df = pd.DataFrame(episode_info)
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/bestEpisode-{}_Info_normRew_{}.csv".format(episode,
                                                                                    int(time.time() * 1000)))
        df = pd.DataFrame(episode_reward)
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/bestEpisode-{}_rewards_normRew_{}.csv".format(episode,
                                                                                       int(
                                                                                           time.time() * 1000)))
        df = None
        env.matlabEng.eval("saveSDIData('{}', {})".format((f"{pythonFilePath}/logs/TD3_cutOut_Scenario/trial_1/"), episode),
                           nargout=0)
        best_score = np.mean(episode_reward)  #(np.sum(episode_reward) / (num_steps))
        best_path_actions = episode_actions
        best_score_episodes_taken = episode
        best_score_info = episode_info
        df = pd.DataFrame(
            agent.memory.state_buffer[:agent.memory.buffer_counter])
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/best_Statesbuffer_normRew_{}_{}.csv".format(int(time.time() * 1000),
                                                                                     episode))
        df = pd.DataFrame(
            agent.memory.action_buffer[:agent.memory.buffer_counter])
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/best_Actionsbuffer_normRew_{}_{}.csv".format(int(time.time() * 1000),
                                                                                      episode))
        df = pd.DataFrame(
            agent.memory.reward_buffer[:agent.memory.buffer_counter])
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/best_rewardsbuffer_normRew_{}_{}.csv".format(int(time.time() * 1000),
                                                                                      episode))
        df = pd.DataFrame(
            agent.memory.next_state_buffer[:agent.memory.buffer_counter])
        df.to_csv(
            r"logs/TD3_cutOut_Scenario/trial_1/best_nxtStatebuffer_normRew_{}_{}.csv".format(int(time.time() * 1000),
                                                                                       episode))
        agent.save_bestmodels("{}_{}".format(int(time.time() * 1000), episode))
        df = None
