import os
import torch
import torch.nn as nn
import torch.nn.functional as F 
import torch.autograd
from torch.autograd import Variable

class Critic(nn.Module):
    def __init__(self, input_size, hidden_size, output_size, droprob, name, chkpt_dir='tmp/ddpg'):
        
        seed = 0
        torch.manual_seed(seed)            # CPU seed
        torch.cuda.manual_seed(seed)       # current GPU seed
        torch.cuda.manual_seed_all(seed)   # all GPU seed
        
        super(Critic, self).__init__()
        self.linear1 = nn.Linear(input_size, hidden_size)
        self.linear2 = nn.Linear(hidden_size, hidden_size)
        self.linear3 = nn.Linear(hidden_size, output_size)
        
        self.dropout = torch.nn.Dropout(p=droprob) #p probability of an element to be zeroed
        
        self.checkpoint_dir = chkpt_dir
        self.checkpoint_file = os.path.join(self.checkpoint_dir, name+'_ddpg')
        
        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        self.to(self.device)

    def forward(self, state, action):
        """
        Params state and actions are torch tensors
        """
        x = torch.cat([state, action], 1) 
        x = F.relu(self.linear1(x))
        x = self.dropout(x)
        x = F.relu(self.linear2(x))
        x = self.dropout(x)
        x = self.linear3(x)

        return x
    
    def save_checkpoint(self):
        #print('... saving checkpoint ...')
        torch.save(self.state_dict(), self.checkpoint_file)
    
    def load_checkpoint(self):
        #print('... loading checkpoint ...')
        self.load_state_dict(torch.load(self.checkpoint_file))

class Actor(nn.Module):
    def __init__(self, input_size, hidden_size, output_size, droprob, name, learning_rate = 3e-4, chkpt_dir='tmp/ddpg'):
        
        seed = 0
        torch.manual_seed(seed)            # CPU seed
        torch.cuda.manual_seed(seed)       # current GPU seed
        torch.cuda.manual_seed_all(seed)   # all GPU seed
        
        super(Actor, self).__init__()
        self.linear1 = nn.Linear(input_size, hidden_size)
        self.linear2 = nn.Linear(hidden_size, hidden_size)
        self.linear3 = nn.Linear(hidden_size, output_size)
        
        self.dropout = torch.nn.Dropout(p=droprob)
        
        self.checkpoint_dir = chkpt_dir
        self.checkpoint_file = os.path.join(self.checkpoint_dir, name+'_ddpg')
        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        self.to(self.device)
        
    def forward(self, state):
        """
        Param state is a torch tensor
        """
        x = F.relu(self.linear1(state))
        x = self.dropout(x)
        x = F.relu(self.linear2(x))
        x = self.dropout(x)
        x = torch.tanh(self.linear3(x))

        return x
    
    def save_checkpoint(self):
        #print('... saving checkpoint ...')
        torch.save(self.state_dict(), self.checkpoint_file)
    
    def load_checkpoint(self):
        #print('... loading checkpoint ...')
        self.load_state_dict(torch.load(self.checkpoint_file))

